#[macro_use]
extern crate log;

#[macro_use]
extern crate snafu;

extern crate env_logger;
extern crate ommui_data;
extern crate ommui_file_loading;
extern crate structopt;
extern crate xio_instructionset;
extern crate xio_jobset;

use ommui_data::xio_job_to_blockdiag::WriteJob;
use ommui_file_loading::PathLoad;
use snafu::{ErrorCompat, ResultExt};
use structopt::StructOpt;
use xio_instructionset::InstructionMap;
use xio_jobset::Job;

#[derive(Debug, Snafu)]
enum Error {
    #[snafu(display("{}", source))]
    OmmuiFileLoading { source: ommui_file_loading::Error },
    #[snafu(display("{}", source))]
    OmmuiData { source: ommui_data::Error },
}

type Result<T, E = Error> = std::result::Result<T, E>;

#[derive(StructOpt, Debug)]
#[structopt(
    name = "xiojob2blockdiag",
    about = "Command-line tool to export XIO jobs to blockdiag"
)]
struct Opt {
    #[structopt(long = "verbose", short = "v")]
    verbose: bool,
    #[structopt(
        long = "instructionset",
        default_value = "/usr/share/xio-schema-definition/instructionset/1"
    )]
    instructionset: String,
    #[structopt(name = "job")]
    job: String,
}

fn run(opt: &Opt) -> Result<()> {
    let instructions = InstructionMap::load_from_path(&format!(
        "{}/instructionset.json",
        &opt.instructionset
    ))
    .context(OmmuiFileLoading)?;
    let job = Job::load_from_path(&opt.job).context(OmmuiFileLoading)?;

    std::io::stdout()
        .write_job(&job, &instructions)
        .context(OmmuiData)?;
    Ok(())
}

fn main() {
    let opt = Opt::from_args();

    let mut builder = env_logger::Builder::new();
    builder.parse_filters("xio_job_to_blockdiag=info");
    builder.parse_filters("xiojob2blockdiag=info");
    builder.parse_filters("xio_jobset=info");
    if let Ok(rust_log) = std::env::var("RUST_LOG") {
        builder.parse_filters(&rust_log);
    }
    if opt.verbose {
        builder.parse_filters("debug");
    }
    builder.target(env_logger::Target::Stderr);
    builder.init();

    if let Err(s) = run(&opt) {
        warn!("{:#?}", s);
        if let Some(backtrace) = ErrorCompat::backtrace(&s) {
            warn!("{}", backtrace);
        }
        std::process::exit(1);
    }
}
